/* $Id: ast.c,v 1.10 2023/03/29 20:46:56 leavens Exp $ */
#include <stdlib.h>
#include "utilities.h"
#include "ast.h"

#define NUMOPERS 8
static const char *opstrs[NUMOPERS] = {
    "==", "!=", "<", "<=", "+", "-", "*", "/"};
// Return a string that is the syntax of the given oper
extern const char *oper2str(oper op)
{
    return opstrs[op];
}

// Return a (pointer to a) fresh AST
// and fill in its file_location with the given file name (fn),
// line number (ln) and column number (col).
// Also initializes the next pointer to NULL.
// If there is no space to allocate an AST node,
// print an error on stderr and exit with a failure code.
static AST *ast_allocate(const char *fn, unsigned int ln, unsigned int col)
{
    AST *ret = (AST *) malloc(sizeof(AST));
    if (ret == NULL) {
	bail_with_error("No space to create const_def AST!");
    }
    ret->file_loc.filename = fn;
    ret->file_loc.line = ln;
    ret->file_loc.column = col;
    ret->next = NULL;
    return ret;
}

// Return a (pointer to a) fresh AST for a program, whose first token
// starts in the given file (fn), line (ln), and column (col),
// and which contains the given AST_list for var-decls (vds) and an AST
// for the statement (stmt).
AST *ast_program(const char *fn, unsigned int ln, unsigned int col,
		 AST_list vds, AST *stmt)
{
    AST *ret = ast_allocate(fn, ln, col);
    ret->type_tag = program_ast;
    ret->data.program.vds = vds;
    ret->data.program.stmt = stmt;
    return ret;
}

// Return a (pointer to a) fresh AST for a var declaration
// with var_type vt and name ident.
AST *ast_var_decl(token t, var_type vt, const char *ident)
{
    AST *ret = ast_allocate(t.filename, t.line, t.column);
    ret->type_tag = var_decl_ast;
    ret->data.var_decl.vt = vt;
    ret->data.var_decl.name = ident;
    return ret;
}

// Return a (pointer to a) fresh AST for an assignment statement
// with the given name and expression (exp).
AST *ast_assign_stmt(token t, const char *name, AST *exp)
{
    AST *ret = ast_allocate(t.filename, t.line, t.column);
    AST *ident = ast_allocate(t.filename, t.line, t.column);
    ident->type_tag = ident_ast;
    ident->data.ident.name = name;
    id_attrs *attrs = id_attrs_start(token2file_loc(t));
    ident->data.ident.idu = id_use_create(attrs, 0);
    ret->type_tag = assign_ast;
    ret->data.assign_stmt.ident = ident;
    ret->data.assign_stmt.exp = exp;
    return ret;
}

// Return a (pointer to a) fresh AST for a begin-statement
// with statments AST stmts.
AST *ast_begin_stmt(token t, AST_list vds, AST_list stmts)
{
    AST *ret = ast_allocate(t.filename, t.line, t.column);
    ret->type_tag = begin_ast;
    ret->data.begin_stmt.vds = vds;
    ret->data.begin_stmt.stmts = stmts;
    return ret;
}

// Return a (pointer to a) fresh AST for an if-statement
// with expression AST exp and statement stmt
AST *ast_if_stmt(token t, AST *exp, AST *stmt)
{
    AST *ret = ast_allocate(t.filename, t.line, t.column);
    ret->type_tag = if_ast;
    ret->data.if_stmt.exp = exp;
    ret->data.if_stmt.stmt = stmt;
    return ret;
}

// Return a (pointer to a) fresh AST for a read-statement
// with variable identifier name
AST *ast_read_stmt(token t, const char *name)
{
    AST *ret = ast_allocate(t.filename, t.line, t.column);
    AST *ident = ast_allocate(t.filename, t.line, t.column);
    ident->type_tag = ident_ast;
    ident->data.ident.name = name;
    id_attrs *attrs = id_attrs_start(token2file_loc(t));
    ident->data.ident.idu = id_use_create(attrs, 0);
    ret->type_tag = read_ast;
    ret->data.read_stmt.ident = ident;
    return ret;
}

// Return a (pointer to a) fresh AST for a write-statement
// with expression AST exp
AST *ast_write_stmt(token t, AST *exp)
{
    AST *ret = ast_allocate(t.filename, t.line, t.column);
    ret->type_tag = write_ast;
    ret->data.write_stmt.exp = exp;
    return ret;
}

// Return a (pointer to a) fresh AST for a pair of a binary operator, op,
// and a (right) expression e2
AST *ast_op_expr(token t, oper op, AST *e2)
{
    AST *ret = ast_allocate(t.filename, t.line, t.column);
    ret->data.op_expr.op = op;
    ret->data.op_expr.exp = e2;
    return ret;
}

// Return a (pointer to a) fresh AST for a binary expression
// with left expresion AST e1, binary artihmetic operator arith_op,
// and right expression AST e2.
AST *ast_bin_expr(token t, AST *e1, oper op, AST *e2)
{
    AST *ret = ast_allocate(t.filename, t.line, t.column);
    ret->type_tag = bin_expr_ast;
    ret->data.bin_expr.leftexp = e1;
    ret->data.bin_expr.op = op;
    ret->data.bin_expr.rightexp = e2;
    return ret;
}

// Return a (pointer to a) fresh AST for an identref expression
// with the given name.
AST *ast_ident(token t, const char *name)
{
    AST *ret = ast_allocate(t.filename, t.line, t.column);
    ret->type_tag = ident_ast;
    ret->data.ident.name = name;
    id_attrs *attrs = id_attrs_start(token2file_loc(t));
    ret->data.ident.idu = id_use_create(attrs, 0);
    return ret;
}

// Return a (pointer to a) fresh AST for an (signed) number expression
// with the given value
AST *ast_number(token t, float value)
{
    AST *ret = ast_allocate(t.filename, t.line, t.column);
    ret->type_tag = number_ast;
    ret->data.number.value = word_from_float(value);
    return ret;
}

// Return a (pointer to a) fresh AST for logical negation expression
// with the given expression AST exp
AST *ast_not_expr(token t, AST *exp)
{
    AST *ret = ast_allocate(t.filename, t.line, t.column);
    ret->type_tag = not_expr_ast;
    ret->data.not_expr.exp = exp;
    return ret;
}


// Return an AST list that is empty
AST_list ast_list_empty_list()
{
    return NULL;
}

// Return an AST list consisting of just the given AST node (ast)
AST_list ast_list_singleton(AST *ast)
{
    return ast;
}

// Return true just when lst is an empty list (and false otherwise)
bool ast_list_is_empty(AST_list lst)
{
    return lst == NULL;
}

// Requires lst != NULL
// Return the first element in an AST_list
AST *ast_list_first(AST_list lst)
{
    // each node is an AST itself, so the first element is lst
    return lst;
}

// Requires lst != NULL
// Return the rest of the AST_list (which is null if it is empty)
AST_list ast_list_rest(AST_list lst)
{
    return lst->next;
}


// Return the last element in the AST list lst.
// The result is only NULL if ast_list_is_empty(lst);
AST_list ast_list_last_elem(AST_list lst)
{
    AST *prev = NULL;
    while (!ast_list_is_empty(lst)) {
	prev = lst;
	lst = ast_list_rest(lst);
    }
    // here ast_list_is_empty(lst)
    return prev;
}

// Requires: !ast_list_is_empty(lst) and ast_list_is_empty(ast_list_rest(lst))
// Make newtail the tail of the AST_list starting at lst
void ast_list_splice(AST_list lst, AST_list newtail)
{
    lst->next = newtail;
}

// Return the number of elements in the AST list lst.
extern int ast_list_size(AST_list lst)
{
    int ret = 0;
    while (!ast_list_is_empty(lst)) {
	ret++;
	lst = ast_list_rest(lst);
    }
    return ret;
}
