/* $Id: ast.h,v 1.11 2023/03/29 20:46:56 leavens Exp $ */
#ifndef _AST_H
#define _AST_H
#include <stdbool.h>
#include "token.h"
#include "file_location.h"
#include "type_attrs.h"
#include "machine_types.h"
#include "id_use.h"

// types of ASTs (type tags)
typedef enum {
    program_ast, var_decl_ast,
    assign_ast, begin_ast,
    if_ast, read_ast, write_ast,
    op_expr_ast, bin_expr_ast, 
    ident_ast, number_ast, not_expr_ast
} AST_type;

// forward declaration, so can use the type AST* below
typedef struct AST_s AST;
// lists of ASTs
typedef AST *AST_list;

// The following types for structs named N_t
// are used in the declaration of the AST_s struct below.
// The struct N_t is the type of information kept in the AST
// that is related to the nonterminal N in the abstract syntax.
// In addition there is an enum types declared before AST_s,
// for binary operators used in expressions (oper).

// P ::= { VD } S
typedef struct {
    AST_list vds;
    AST *stmt;
} program_t;

// VD ::= float x | bool x
typedef struct {
    var_type vt;
    const char *name;
} var_decl_t;

// S ::= assign x E
typedef struct {
    AST *ident;
    AST *exp;
} assign_t;

// S ::= begin { VD } { S }
typedef struct {
    AST_list vds;
    AST_list stmts;
} begin_t;

// S ::= if E S
typedef struct {
    AST *exp;
    AST *stmt;
} if_t;

// S ::= read x
typedef struct {
    AST *ident;
} read_t;

// S ::= write E
typedef struct {
    AST *exp;
} write_t;

// o ::= == | != | < | <= | + | - | * | /
typedef enum {eqeqop, neqop, ltop, leqop, plusop, minusop, multop, divop} oper;
// Return a string that is the syntax of the given oper
extern const char *oper2str(oper op);

// E ::= E1 o E2
typedef struct {
    AST *leftexp;
    oper op;
    AST *rightexp;
} bin_expr_t;

// E ::= o E
// The following is used for pairs of operator and an expression
// (when parsing terms and factors),
// but it is converted into some other kind of AST eventually
typedef struct {
    oper op;
    AST *exp;
} op_expr_t;

// E ::= x
typedef struct {
    // name of a constant or variable
    const char *name;
    // set during static analysis,
    // includes info for lexical addr
    id_use *idu;
} ident_t;

// E ::= n
typedef struct {
    word value;
} number_t;

// E ::= !E
typedef struct {
    AST *exp;
} not_expr_t;

// The actual AST definition:
typedef struct AST_s {
    file_location file_loc;
    AST_list next;  // for lists
    AST_type type_tag;
    union AST_u {
	program_t program;
	var_decl_t var_decl;
	assign_t assign_stmt;
	begin_t begin_stmt;
	if_t if_stmt;
	read_t read_stmt;
	write_t write_stmt;
	bin_expr_t bin_expr;
	op_expr_t op_expr;  // only used by the parser
	ident_t ident;
	number_t number;
	not_expr_t not_expr;
    } data;
} AST;

// Return a (pointer to a) fresh AST for a program, whose first token
// starts in the given file (fn), line (ln), and column (col),
// and which contains the given ASTs for const-decls (cds), var-decls (vds)
// and statement (stmt).
extern AST *ast_program(const char *fn, unsigned int ln, unsigned int col,
		 AST_list vds, AST *stmt);

// Return a (pointer to a) fresh AST for a var declaration of type vt
// with name ident, which starts at the token t
extern AST *ast_var_decl(token t, var_type vt, const char *ident);

// Return a (pointer to a) fresh AST for an assignment statement
// with name ident and expression AST exp.
extern AST *ast_assign_stmt(token t, const char *ident, AST *exp);

// Return a (pointer to a) fresh AST for a begin-statement
// with var-decl ASTs vds and statments AST stmts.
extern AST *ast_begin_stmt(token t, AST_list vds, AST_list stmts);

// Return a (pointer to a) fresh AST for an if-statement
// with expression AST exp and statement stmt
extern AST *ast_if_stmt(token t, AST *exp, AST *stmt);

// Return a (pointer to a) fresh AST for a read-statement
// with variable identifier name
extern AST *ast_read_stmt(token t, const char *name);

// Return a (pointer to a) fresh AST for a write-statement
// with expression AST exp
extern AST *ast_write_stmt(token t, AST *exp);

// Return a (pointer to a) fresh AST for a pair of a binary operator, op,
// and a (right) expression e2
extern AST *ast_op_expr(token t, oper op, AST *e2);

// Return a (pointer to a) fresh AST for a binary expression
// with left expresion AST e1, operator op,
// and right expression AST e2.
extern AST *ast_bin_expr(token t, AST *e1, oper op, AST *e2);

// Return a (pointer to a) fresh AST for an ident expression
// with the given name.
extern AST *ast_ident(token t, const char *name);

// Return a (pointer to a) fresh AST for a (signed) number expression
// with the given value
extern AST *ast_number(token t, float value);

// Return a (pointer to a) fresh AST for logical negation expression
// with the given expression AST exp
extern AST *ast_not_expr(token t, AST *exp);


// Return an AST list that is empty
extern AST_list ast_list_empty_list();

// Return an AST list consisting of just the given AST node (ast)
extern AST_list ast_list_singleton(AST *ast);

// Return true just when lst is an empty list (and false otherwise)
extern bool ast_list_is_empty(AST_list lst);

// Requires: !ast_list_is_empty(lst)
// Return the first element in an AST_list
extern AST *ast_list_first(AST_list lst);

// Requires: !ast_list_is_empty(lst)
// Return the rest of the AST_list (which is null if it is empty)
extern AST_list ast_list_rest(AST_list lst);

// Requires: !ast_list_is_empty(lst) and ast_list_is_empty(ast_list_rest(lst))
// Make newtail the tail of the AST_list starting at lst
extern void ast_list_splice(AST_list lst, AST_list newtail);

// Return the last element in the AST list lst.
// The result is only NULL if ast_list_is_empty(lst);
extern AST_list ast_list_last_elem(AST_list lst);

// Return the number of elements in the AST list lst.
extern int ast_list_size(AST_list lst);

#endif
