/* $Id: code.c,v 1.9 2023/03/30 19:56:21 leavens Exp $ */
#include <stdlib.h>
#include "utilities.h"
#include "lexical_address.h"
#include "code.h"

// Code creation functions below

typedef enum {
     NOP, LIT, RTN, CAL, POP, PSI, LOD, STO, INC, JMP,
     JPC, CHO, CHI, HLT, NDB, NEG, ADD, SUB, MUL, DIV,
     RND, EQL, NEQ, LSS, LEQ, GTR, GEQ, PSP, PBP, PPC,
     JMI, RBP
} opcode;

// Return a fresh code struct, with next pointer NULL
// containing the instruction with opcode op and the given m parameter.
// If there is not enough space, bail with an error,
// so this will never return NULL.
static code *code_create(opcode op, word m)
{
    code *ret = (code *)malloc(sizeof(code));
    if (ret == NULL) {
	bail_with_error("Not enough space to allocate a code struct!");
    }
    ret->next = NULL;
    ret->instr.op = op;
    ret->instr.m = m;
    return ret;
}

// no-op
code *code_nop()
{
    return code_create(NOP, word_from_int(0));
}

// literal push
code *code_lit(float f)
{
    return code_create(LIT, word_from_float(f));
}

// return from a subroutine
code *code_rtn()
{
    return code_create(RTN, word_from_int(0));
}

// call the procedure at code index addr
code *code_cal(address addr)
{
    code *ret = code_create(CAL, word_from_int(addr));
    return ret;
}

// pop the stack
code *code_pop()
{
    return code_create(POP, word_from_int(0));
}

// push the element whose address is on the top of the stack
code *code_psi()
{
    return code_create(PSI, word_from_int(0));
}

// load the value at the address given at the top of the stack
// + offset o + BLOCK_LINKS_SIZE
code *code_lod(int o)
{
    return code_create(LOD, word_from_int(BLOCK_LINKS_SIZE+o));
}

// store the top of the stack's value into the address at the
// stack[SP-2]+o+BLOCK_LINK_SIZE, and then pop twice
code *code_sto(int o)
{
    code *ret = code_create(STO,
		        word_from_int(BLOCK_LINKS_SIZE+o));
    return ret;
}

// allocate m locals on the stack
code *code_inc(unsigned int m)
{
    return code_create(INC, word_from_int(m));
}

// jump relative to the current instruction's address plus offset
code *code_jmp(int offset)
{
    return code_create(JMP, word_from_int(offset));
}

// jump conditionally, relative to current instruction's address by
// the given offset if top of the stack is not 0
code *code_jpc(int offset)
{
    return code_create(JPC, word_from_int(offset));
}

// output char on top of the stack
code *code_cho()
{
    return code_create(CHO, word_from_int(0));
}

// read char from stdin and push it onto the stack
code *code_chi()
{
    return code_create(CHI, word_from_int(0));
}

// halt execution
code *code_hlt()
{
    return code_create(HLT, word_from_int(0));
}

// stop printing debugging output
code *code_ndb()
{
    return code_create(NDB, word_from_int(0));
}

// negate value at top of stack
code *code_neg()
{
    return code_create(NEG, word_from_int(0));
}

// add top two elements in the stack
code *code_add()
{
    return code_create(ADD, word_from_int(0));
}

// subtract, top gets stack[SP-2] - stack[SP-1]
code *code_sub()
{
    return code_create(SUB, word_from_int(0));
}

// multiply the top two elements in the stack
code *code_mul()
{
    return code_create(MUL, word_from_int(0));
}

// divide, top gets stack[SP-2] / stack[SP-1]
code *code_div()
{
    return code_create(DIV, word_from_int(0));
}

// round, top gets round(stack[SP-1])
code *code_rnd()
{
    return code_create(RND, word_from_int(0));
}

// equal test
code *code_eql()
{
    return code_create(EQL, word_from_int(0));
}

// not equal test
code *code_neq()
{
    return code_create(NEQ, word_from_int(0));
}

// strictly less than test
code *code_lss()
{
    return code_create(LSS, word_from_int(0));
}

// less than or equal to test
code *code_leq()
{
    return code_create(LEQ, word_from_int(0));
}

// strictly greater than test
code *code_gtr()
{
    return code_create(GTR, word_from_int(0));
}

// greater than or equal to test
code *code_geq()
{
    return code_create(GEQ, word_from_int(0));
}

// push SP on top of the stack
code *code_psp()
{
    return code_create(PSP, word_from_int(0));
}

// push BP on top of the stack
code *code_pbp()
{
    return code_create(PBP, word_from_int(0));
}

// push PC on top of the stack
code *code_ppc()
{
    return code_create(PPC, word_from_int(0));
}

// jump to the address on top of stack
code *code_jmi()
{
    return code_create(JMI, word_from_int(0));
}

// restore BP from top of stack
code *code_rbp()
{
    return code_create(RBP, word_from_int(0));
}


// Sequence manipulation functions below

// Return an empty code_seq
code_seq code_seq_empty()
{
    return NULL;
}

// Return a code_seq containing just the given code
code_seq code_seq_singleton(code *c)
{
    return c;
}


// Is seq empty?
bool code_seq_is_empty(code_seq seq)
{
    return seq == NULL;
}

// Requires: !code_seq_is_empty(seq)
// Return the first element of the given code sequence, seq
code *code_seq_first(code_seq seq)
{
    return seq;
}

// Requires: !code_seq_is_empty(seq)
// Return the rest of the given sequence, seq
code_seq code_seq_rest(code_seq seq)
{
    return seq->next;
}

// Return the size (number of instructions/words) in seq
unsigned int code_seq_size(code_seq seq)
{
    unsigned int ret = 0;
    while (!code_seq_is_empty(seq)) {
	ret++;
	seq = code_seq_rest(seq);
    }
    return ret;
}

// Requires: !code_seq_is_empty(seq)
// Return the last element in the given sequence
code *code_seq_last_elem(code_seq seq)
{
    code *ret = seq;
    while (!code_seq_is_empty(seq)) {
	ret = seq;
	seq = code_seq_rest(seq);
    }
    return ret;
}

// Requires: c != NULL
// Add the given code *c to the end of the seq
// and return the seq, which has been modified if it was not empty
code_seq code_seq_add_to_end(code_seq seq, code *c)
{
    if (code_seq_is_empty(seq)) {
	return code_seq_singleton(c);
    }
    // assert(!code_seq_is_empty(seq));
    code *last = code_seq_last_elem(seq);	
    last->next = c;
    c->next = NULL;
    return seq;
}

// Concatenate the given code sequences in order first s1 then s2
// This may modify the sequence s1 if both s1 and s2 are not empty
code_seq code_seq_concat(code_seq s1, code_seq s2)
{
    if (code_seq_is_empty(s1)) {
	return s2;
    } else if (code_seq_is_empty(s2)) {
	return s1;
    } else {
	code *last = code_seq_last_elem(s1);
	last->next = s2;
	return s1;
    }
}

// Requires: out is open for writing
// print the instructions in the code_seq to out in debugging format
void code_seq_debug_print(FILE *out, code_seq seq)
{
    instruction_heading_print(out);
    int addr = 0;
    while(!code_seq_is_empty(seq)) {
	instruction_debug_print_with_addr(out, addr,
					  code_seq_first(seq)->instr);
	seq = code_seq_rest(seq);
	addr++;
    }
}

// Requires: out is open for writing
// print the instructions in seq, using the VM's input format, to out
void code_seq_vm_print(FILE *out, code_seq seq)
{
    while(!code_seq_is_empty(seq)) {
	instruction_vm_format_print(out, code_seq_first(seq)->instr);
	seq = code_seq_rest(seq);
    }
}

// Return a code sequence that will put the address that corresponds to the
// frame pointer for the given number of scopes outward on top of the stack
code_seq code_compute_fp(unsigned int levelsOut)
{
    code_seq ret = code_seq_singleton(code_pbp()); // bp of current stack frame
    while (levelsOut > 0) {
	// add in code to follow each needed static link
	ret = code_seq_add_to_end(ret, code_psi(0));
	levelsOut--;
    }
    // now ret holds the code to put the address of the frame for *la
    // on the top of the stack
    return ret;
}

// Requires: la != NULL
// Return a code sequence that will put the value
// of the given lexical address on the top of the stack
code_seq code_load_from_lexical_address(lexical_address *la)
{
    
    code_seq ret = code_compute_fp(la->levelsOutward);
    code_seq_add_to_end(ret,
			code_seq_singleton(code_lod(la->offsetInAR)));
    return ret;
}

