/* $Id: instruction.c,v 1.3 2023/03/30 02:14:01 leavens Exp $ */
#include <stdio.h>
#include <stdbool.h>
#include "instruction.h"
#include "utilities.h"

// one more than the highest op code, to allow for 0
#define NUM_OPCODES 32

static const char *opcodes[NUM_OPCODES] =
    {"NOP", "LIT", "RTN", "CAL", "POP",
     "PSI", "LOD", "STO", "INC", "JMP",
     "JPC", "CHO", "CHI", "HLT", "NDB",
     "NEG", "ADD", "SUB", "MUL", "DIV",
     "RND", "EQL", "NEQ", "LSS", "LEQ",
     "GTR", "GEQ", "PSP", "PBP", "PPC",
     "JMI", "RBP"};

// Is the argument a legal op code for the machine?
bool legal_op_code(int op)
{
    return 0 <= op && op < NUM_OPCODES;
}

// return the mnemonic for the given op code
const char * mnemonic(int op)
{
    if (!legal_op_code(op)) {
	bail_with_error("Illegal opcode: %d", op);
    }
    return opcodes[op];
}

bool stop_reading = false;

// read a single instruction (all on one line) from the file in and return it
// sets stop_reading to true if there is an error or EOF detected
instruction read_instruction(FILE *in)
{
    instruction instr;
    int num_read = fscanf(in, "%d", &instr.op);
    if (!legal_op_code(instr.op) || num_read < 1) {
	stop_reading = true;
    }
    if (instr.op != 1) {
	instr.m.type_tag = int_type;
	num_read = fscanf(in, " %d\n", &instr.m.data.i);
    } else {
	instr.m.type_tag = float_type;
	num_read = fscanf(in, " %f\n", &instr.m.data.f);
    }
    if (num_read < 1) {
	stop_reading = true;
    }
    return instr;
}

// Requires: out is open and writable
// print the header of the output table to out
void instruction_heading_print(FILE *out) {
    fprintf(out, "%-5s %-5s %-5s\n", "Addr", "OP", "M");
}

// Requires: out is an open FILE
// print the instruction with mnemonic (on one line)
// followed by a newline character.
void instruction_debug_format_print(FILE *out, instruction instr)
{
    if (!legal_op_code(instr.op)) {
	    bail_with_error("Illegal opcode passed to instruction_debug_format_print %d",
			    instr.op);
    }
    fprintf(out, "%-5s ", mnemonic(instr.op));
    word_print(out, instr.m);
    fprintf(out, "\n");
}

// Requires: out is open for writing
// print to out the given instruction, which is found at the given address
void instruction_debug_print_with_addr(FILE *out, int addr, instruction instr)
{
    fprintf(out, "%-5d ", addr);
    instruction_debug_format_print(out, instr);
}

// Requires: out is an open FILE
// print the instruction in the VM's input format,
// followed by a newline character.
void instruction_vm_format_print(FILE *out, instruction instr)
{
    if (!legal_op_code(instr.op)) {
	    bail_with_error("Illegal opcode passed to instruction_vm_format_print %d",
			    instr.op);
    }
    fprintf(out, "%-5d ", instr.op);
    word_print(out, instr.m);
    fprintf(out, "\n");
}
