/* $Id: lexer.c,v 1.9 2023/04/06 21:38:31 leavens Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include "token.h"
#include "utilities.h"
#include "lexer.h"
#include "reserved.h"
#include "machine_types.h"

// The input file
static FILE *input_file = NULL;
// The input file's name
static const char *filename = NULL;
// Is this token stream done (past EOF or error)?
static bool done = true;
// the line of the next token
static unsigned int line;
// the column of the next token
static unsigned int column;

// Check the lexer's invariant
static void lexer_okay()
{
    assert(done == (input_file == NULL));
    assert(done == (filename == NULL));
}

// Initialize the lexer (i.e., its data structures)
static void lexer_initialize()
{
    filename = NULL;
    input_file = NULL;
    done = true;
    line = 1;
    column = 1;
    reserved_initialize();
}

// Requires: fname != NULL
// Requires: fname is the name of a readable file
// Initialize the lexer and start it reading
// from the given file name
void lexer_open(const char *fname)
{
    lexer_initialize();
    input_file = fopen(fname, "r");
    if (input_file == NULL) {
	bail_with_error("Cannot open %s", fname);
    }
    filename = fname;
    done = feof(input_file);
    if (done) {
	input_file = NULL;
	filename = NULL;
    }
    lexer_okay();
}

// Close the file the lexer is working on
// and make this lexer be done
void lexer_close()
{
    lexer_okay();
    if (input_file != NULL) {
	int rc = fclose(input_file);
	if (rc == EOF) {
	    bail_with_error("Cannot close %s!", filename);
	}
    }
    input_file = NULL;
    filename = NULL;
    done = true;
    lexer_okay();
}

// Is the lexer's token stream finished
// (either past EOF or not open)?
bool lexer_done()
{
    return done;
}

// the previous value of column,
// for use in lexer_ungetchar
static unsigned int last_column = 0;

// Requires: input_file is readable
// Return the next char in the input
// updating line and column as appropriate
// update last_column to the old value of column
static char lexer_getchar()
{
    char c = getc(input_file);
    last_column = column;
    if (c == '\n') {
	line++;
	column=1;
    } else {
	column++;
    }
    return c;
}

// Requires: input_file is readable
// Put c back into the input file stream
// to be read again
static void lexer_ungetchar(char c)
{
    column = last_column;
    if (c == '\n') {
	line--;
    }
    if (c != EOF) {
	ungetc(c, input_file);
    }
}

// forward declarations of lexical functions
static void lexer_consume_ignored();
static token lexer_ident(char c, token t);
static token lexer_number(char c, token t);
static token lexer_starts_eq(char c, token t);
static token lexer_starts_exclaim(char c, token t);
static token lexer_starts_less(char c, token t);
static bool issign(char c);

// Requires: !lexer_done()
// Return the next token in the input file,
// advancing in the input
token lexer_next()
{
    token t;
    t.filename = filename;
    t.typ = eofsym;
    t.text = NULL;
    t.value = word_from_int(0);

    lexer_consume_ignored();

    t.line = line;
    t.column = column;

    char c = lexer_getchar();
    
    // since we consumed all the whitespace
    // c should not be a kind of space character
    assert(!isspace(c));
    
    if (c == EOF) {
	t.typ = eofsym;
	t.text = NULL;
	filename = NULL;
	input_file = NULL;
	done = true;
	return t;
    }
    if (isalpha(c) || c == '_') {
	return lexer_ident(c, t);
    } else if (isdigit(c) || issign(c) || c == '.') {
	return lexer_number(c, t);
    } else {
	// leave room in t.text for 2 chars of text
	t.text = malloc(3 * sizeof(char));
	if (t.text == NULL) {
	    bail_with_error("Cannot allocate space for token text!");
	}
	t.text[0] = c;
	t.text[1] = '\0';
	switch (c) {
	case ';':
	    t.typ = semisym;
	    break;
	case ',':
	    t.typ = commasym;
	    break;
	case '=':
	    return lexer_starts_eq(c, t);
	    break;
	case '(':
	    t.typ = lparensym;
	    break;
	case ')':
	    t.typ = rparensym;
	    break;
	case '{':
	    t.typ = lcurlysym;
	    break;
	case '}':
	    t.typ = rcurlysym;
	    break;
	case '<':
	    return lexer_starts_less(c, t);
	    break;
	case '!':
	    return lexer_starts_exclaim(c, t);
	    break;
	case '*':
	    t.typ = multsym;
	    break;
	case '/':
	    t.typ = divsym;
	    break;
	default:
	    lexical_error(filename, line, column,
			  "Illegal character '%c' (0%o)",
			   c, c);
	    break;
	}
	return t;
    }
}

// Requires: !lexer_done()
// Return the name of the current file
const char *lexer_filename()
{
    if (lexer_done()) {
	bail_with_error("Asking for file name of done lexer!");
    }
    return filename;
}

// Requires: !lexer_done()
// Return the line number of the next token
unsigned int lexer_line()
{
    if (lexer_done()) {
	bail_with_error("Asking for line of done lexer!");
    }
    return line;
}

// Requires: !lexer_done()
// Return the column number of the next token
unsigned int lexer_column()
{
    if (lexer_done()) {
	bail_with_error("Asking for column of done lexer!");
    }
    return column;
}

// Requires: input_file is readable
// Advance input_file to the next newline
static void lexer_consume_comment()
{
    char c = lexer_getchar();
    while (c != '\n' && c != EOF) {
	c = lexer_getchar();
    }
    if (c == EOF) {
	lexical_error(filename, line, column,
		      "File ended while reading comment!");
    }
    // assert(c == '\n');
}

// Requires: input_file is readable
// Advance in the input file until
// the next char is the start of a token
// that is not ignored
// (i.e., not whitespace or a comment)
static void lexer_consume_ignored()
{
    char c = lexer_getchar();
    while (isspace(c) || c == '#') {
	if (isspace(c)) {
	    // ignore the whitespace char
	    c = lexer_getchar();
	} else if (c == '#') {
	    lexer_consume_comment();
	    c = lexer_getchar();
	}
    }
    // assert(!isspace(c) && c != '#');
    lexer_ungetchar(c);
}

// Requires: c is a letter
// Return a token for a reserved word
// or an identifier
static token lexer_ident(char c, token t)
{
    char *text = malloc((MAX_IDENT_LENGTH+1)*sizeof(char));
    if (text == NULL) {
	bail_with_error("Cannot allocate space for identifier");
    }
    text[0] = c;
    int n = 1;
    c = lexer_getchar();
    while (isalpha(c) || isdigit(c) || c == '_') {
	if (n >= MAX_IDENT_LENGTH) {
	    text[n] = '\0';
	    lexical_error(filename, line, column,
			  "Identifier starting \"%s\" is too long!",
			  text);
	}
	text[n] = c;
	n++;
	c = lexer_getchar();
    }
    assert(!isalpha(c) && !isdigit(c) && c != '_');
    text[n+1] = '\0';    
    lexer_ungetchar(c);
    t.text = text;
    t.typ = reserved_type(text);
    return t;
}

// 16 decimal digits, + 1 (.) + 1 (sign) + 1 (e or E) + 3 (power) = 22
#define MAX_NUM_LENGTH 31

// accumulate any digits into text, updating variables c and n
static void lexer_accum_digits(char *c, int *n, char text[])
{
    while (isdigit((int)*c)) {
	if (*n >= MAX_NUM_LENGTH) {
	    text[*n] = '\0';
	    bail_with_error(filename, line, column,
			    "Number starting \"%s\" is too long!",
			    text);
	}
	text[*n] = *c;
	(*n)++;
	*c = lexer_getchar();
    }
}

// Requires: c is a digit
// Return a token for a number
static token lexer_number(char c, token t)
{
    assert(issign(c) || isdigit(c) || c == '.');
    char *text = malloc((MAX_NUM_LENGTH+1)*sizeof(char));
    if (text == NULL) {
	bail_with_error("Cannot allocate space for number");
    }
    text[0] = c;
    int n = 1;
    bool isplus = (c == '+');
    if (issign(c)) {
	// consume the sign char
	c = lexer_getchar();
	text[1] = c;
	n++;
	if (!isdigit(c)) {
	    // the token is not a number
	    text[1] = '\0';
	    lexer_ungetchar(c);
	    t.text = text;
	    t.typ = (isplus ? plussym : minussym);
	    return t;
	}
	// assert(isdigit(c));
    }
    // look at the next char
    c = lexer_getchar();
    lexer_accum_digits(&c, &n, text);
    // assert(!isdigit(c));
    if (c == '.') {
	text[n] = c;
	n++;
	c = lexer_getchar();
	lexer_accum_digits(&c, &n, text);
    }
    if (c == 'e' || c == 'E') {
	text[n] = c;
	n++;
	c = lexer_getchar();
	// must have a sign or a digit next
	if (!(issign(c) || isdigit(c))) {
	    lexical_error(filename, line, column,
			  "number's exponent marker must be followed by a sign or a digit, not '%c'!", c);
	} else {
	    char saved = c;
	    text[n] = c;
	    n++;
	    c = lexer_getchar();
	    if (!isdigit(saved)) {
		// need a digit
		if (isdigit(c)) {
		    text[n] = c;
		    n++;
		    c = lexer_getchar();
		} else {
		    lexical_error(filename, line, column,
				  "In an exponent, a sign must be followed by a digit, not '%c'!", c);
		}
	    }
	}
	// then any number of digits
	lexer_accum_digits(&c, &n, text);
    }
    // assert(!isdigit(c));
    text[n] = '\0';
    lexer_ungetchar(c);
    t.text = text;
    float temp;
    int rc = sscanf(text, "%f", &temp);
    t.value = word_from_float(temp);
    // check that the scan was successful
    assert(rc != EOF);
    t.typ = numbersym;
    return t;
}

// Requires: c is a equals-sign character (=)
// Returns the token appropriate for the next char
static token lexer_starts_eq(char c, token t)
{
    assert(c == '=');
    c = lexer_getchar();
    t.text[1] = c;
    t.text[2] = '\0';
    switch (c) {
    case '=':
	t.typ = eqeqsym;
	break;
    default:
	t.text[1] = '\0';
	lexer_ungetchar(c);
	t.typ = becomessym;
	break;
    }
    return t;
}

// Requires: c is a equals-sign character (=)
// Returns the token appropriate for the next char
static token lexer_starts_exclaim(char c, token t)
{
    assert(c == '!');
    c = lexer_getchar();
    t.text[1] = c;
    t.text[2] = '\0';
    switch (c) {
    case '=':
	t.typ = neqsym;
	break;
    default:
	t.text[1] = '\0';
	lexer_ungetchar(c);
	t.typ = exclaimsym;
	break;
    }
    return t;
}


// Requires: c is a less-than character (<)
// Returns the token appropriate for the next char
static token lexer_starts_less(char c, token t)
{
    assert(c == '<');
    c = lexer_getchar();
    t.text[1] = c;
    t.text[2] = '\0';
    switch (c) {
    case '=':
	t.typ = leqsym;
	break;
    default:
	t.text[1] = '\0';
	lexer_ungetchar(c);
	t.typ = ltsym;
	break;
    }
    return t;
}

static bool issign(char c)
{
    return c == '+' || c == '-';
}
