/* $Id: parser.c,v 1.13 2023/03/29 20:46:56 leavens Exp $ */
#include <stdio.h>
#include "lexer.h"
#include "id_attrs.h"
#include "utilities.h"
#include "parserInternal.h"

static token tok;
static unsigned int scope_offset;

// initialize the parser to work on the given file
void parser_open(const char *filename)
{
    lexer_open(filename);
    tok = lexer_next();
    scope_offset = 0;
}

// finish using the parser
void parser_close()
{
    lexer_close();
}

// Advance to the next token
// (if the lexer is not done,
//  otherwise do nothing)
static void advance()
{
    if (!lexer_done()) {
	tok = lexer_next();
    }
}

// Check that the current token is of type tt
// and advance to the next token if so;
// otherwise, produce a syntax error.
static void eat(token_type tt) {
    if (tok.typ == tt) {
	// debug_print("Eating a %s token\n", ttyp2str(tt));
        advance();
    } else {
	token_type expected[1] = {tt};
	parse_error_unexpected(expected, 1, tok);
    }
}

// <program> ::= <var-decls> <stmt> EOF
AST *parseProgram()
{
    AST_list vds = parseVarDecls();
    AST *stmt = parseStmt();
    // cannot be any input left over after the stmt
    eat(eofsym);
    file_location floc;
    if (!ast_list_is_empty(vds)) {
	if (ast_list_first(vds)->type_tag == var_decl_ast) {
	    floc = ast_list_first(vds)->file_loc;
	} else {
	    bail_with_error("Bad AST for var declarations");
	}
    } else {
	floc = stmt->file_loc;
    }
    return ast_program(floc.filename, floc.line, floc.column,
			  vds, stmt);
}

// Requires: !ast_list_is_empty(*head) ==> !ast_list_is_empty(*last).
// Requires: when called head points to the first element of an AST_list
// and last points to the last element in that list.
// Modifies *head, *last;
// Splice the list starting at lst into the AST list starting at *head,
// and make *last point to the last element in the resulting list.
static void add_AST_to_end(AST_list *head, AST_list *last, AST_list lst)
{
    if (ast_list_is_empty(*head)) {
	*head = lst;
	*last = ast_list_last_elem(lst);
    } else {
	// assert(*last != NULL);
	ast_list_splice(*last, lst);
	*last = ast_list_last_elem(lst);
    }
}

// <var-decls> ::= <float-or-bool> <idents> ;
// <float-or-bool>  ::= float | bool
static AST_list parseVarDecls()
{
    AST_list ret = ast_list_empty_list();
    AST_list last = ast_list_empty_list();
    // invariant: !ast_list_is_empty(ret) ==> !ast_list_is_empty(last)
    while (tok.typ == floatsym || tok.typ == boolsym) {
	AST_list vdasts;
	var_type vt;
	if (tok.typ == boolsym) {
	    vt = bool_t;
	    eat(boolsym);
	} else {
	    vt = float_t;
	    eat(floatsym);
	}
	vdasts = parseIdents(vt);
	eat(semisym);
	add_AST_to_end(&ret, &last, vdasts);
    }
    return ret;
}

// <idents> ::= <ident> { <comma-ident> }
static AST_list parseIdents(var_type vt)
{
    token idtok = tok;
    eat(identsym);
    AST_list ret = ast_list_singleton(ast_var_decl(idtok, vt, idtok.text));
    AST_list last = ret;
    while (tok.typ == commasym) {
	eat(commasym);
	token idtok = tok;
	eat(identsym);
	AST *vd = ast_var_decl(idtok, vt, idtok.text);
	add_AST_to_end(&ret, &last, ast_list_singleton(vd));
    }
    return ret;
}

// the kinds of tokens that may begin a statement in FLOAT
#define STMTBEGINTOKS 5
static token_type can_begin_stmt[STMTBEGINTOKS] =
	    {identsym, lcurlysym, ifsym, readsym, writesym};

// can this token begin a statment?
static bool is_stmt_beginning_token(token t)
{
    for (int i = 0; i < STMTBEGINTOKS; i++) {
	if (t.typ == can_begin_stmt[i]) {
	    return true;
	}
    }
    return false;
}

// <stmt> ::= <ident> = <expr> ; | ...
AST *parseStmt()
{
    AST *ret = NULL;
    switch (tok.typ) {
    case identsym:
	ret = parseAssignStmt();
	break;
    case lcurlysym:
	ret = parseBeginStmt();
	break;
    case ifsym:
	ret = parseIfStmt();
	break;
    case readsym:
	ret = parseReadStmt();
	break;
    case writesym:
	ret = parseWriteStmt();
	break;
    default:
	parse_error_unexpected(can_begin_stmt, STMTBEGINTOKS, tok);
    }
    return ret;
}

// <assignment> ::= <ident> = <expr> ;
static AST *parseAssignStmt()
{
    token idtok = tok;
    eat(identsym);
    eat(becomessym);
    AST *exp = parseExpr();
    eat(semisym);
    return ast_assign_stmt(idtok, idtok.text, exp);
}

// <begin-stmt> ::= '{' <stmt> { <stmt> } '}'
static AST_list parseBeginStmt()
{
    token lctok = tok;
    eat(lcurlysym);
    AST_list vds = parseVarDecls();
    AST_list stmts = ast_list_singleton(parseStmt());
    AST_list last = stmts;
    while (is_stmt_beginning_token(tok)) {
	AST *stmt = parseStmt();
	add_AST_to_end(&stmts, &last, ast_list_singleton(stmt));
    }
    eat(rcurlysym);
    AST *ret = ast_begin_stmt(lctok, vds, stmts);
    return ret;
}

// <if-stmt> ::= if ( <condition> ) <stmt>
static AST *parseIfStmt()
{
    token ift = tok;
    eat(ifsym);
    eat(lparensym);
    AST *exp = parseExpr();
    eat(rparensym);
    AST *s = parseStmt();
    return ast_if_stmt(ift, exp, s);
}

// <read-stmt> ::= read <ident> ;
static AST *parseReadStmt()
{
    token rt = tok;
    eat(readsym);
    const char *name = tok.text;
    eat(identsym);
    eat(semisym);
    return ast_read_stmt(rt, name);
}

// <write-stmt> ::= write <expr> ;
static AST *parseWriteStmt()
{
    token wt = tok;
    eat(writesym);
    AST * exp = parseExpr();
    eat(semisym);
    return ast_write_stmt(wt, exp);
}

// Relational operators
#define NUMRELOPS 4
static token_type reloptt[NUMRELOPS]
                  = {eqeqsym, neqsym, ltsym, leqsym};

// Is the given token type one of those that can be used
// as a relational operator?
static bool is_relational_op(token_type tt)
{
    for (int i = 0; i < NUMRELOPS; i++) {
	if (tt == reloptt[i]) {
	    return true;
	}
    }
    return false;
}

// Return the corresponding operator enum for the given token type
static oper tt2oper(token_type tt)
{
    switch (tt) {
    case eqeqsym:
	return eqeqop;
	break;
    case neqsym:
	return neqop;
	break;
    case ltsym:
	return ltop;
	break;
    case leqsym:
	return leqop;
	break;
    case plussym:
	return plusop;
	break;
    case minussym:
	return minusop;
	break;
    case multsym:
	return multop;
	break;
    case divsym:
	return divop;
	break;
    default:
	bail_with_error("Unexpected token type %s (%d) passed to tt2oper!",
			ttyp2str(tt), tt);
	break;
    }
    // the following should never execute!
    return (oper) 0;
}

// Return an AST with the operators (if any) associating to the left
// <expr> ::= <lterm> <rel-op-lterm>
// <rel-op-lterm> ::= <empty> | <rel=op-expr>
AST *parseExpr()
{
    token fst = tok;
    AST *ltrm = parseLTerm();
    AST *exp;
    if (!is_relational_op(tok.typ)) {
	exp = ltrm;
    } else {
	AST *rltrm = parseRelOpExpr();
	exp = ast_bin_expr(fst, ltrm, rltrm->data.op_expr.op,
			   rltrm->data.op_expr.exp);
    }
    return exp;
}

// <lterm> ::= ! <lterm> | <lfactor>
static AST *parseLTerm()
{
    token fst = tok;
    AST *exp;
    if (tok.typ == exclaimsym) {
	eat(exclaimsym);
	AST *negated_exp = parseLTerm();
	exp = ast_not_expr(fst, negated_exp);
    } else {
	exp = parseLFactor();
    }
    return exp;
}

// <rel-op-expr> ::= <rel-op> <lterm>
// This returns an op_expr AST
static AST *parseRelOpExpr()
{
    token op_tok = tok;
    oper op = tt2oper(tok.typ);
    switch (tok.typ) {
    case eqeqsym:
	eat(eqeqsym);
	break;
    case neqsym:
	eat(neqsym);
	break;
    case ltsym:
	eat(ltsym);
	break;
    case leqsym:
	eat(leqsym);
	break;
    default:
	parse_error_unexpected(reloptt, NUMRELOPS, tok);
	break;
    }
    AST *ltrm = parseLTerm();
    return ast_op_expr(op_tok, op, ltrm);
}

// <lfactor> ::= <term> { <add-sub-term> }
static AST *parseLFactor()
{
    token fst = tok;
    AST *exp = parseTerm();
    while (tok.typ == plussym || tok.typ == minussym) {
	AST *rght = parseAddSubTerm();
	exp = ast_bin_expr(fst, exp, rght->data.op_expr.op,
			   rght->data.op_expr.exp);
    }
    return exp;
}

// <add-sub-term> ::= <add-sub> <term>
static AST *parseAddSubTerm()
{
    token opt = tok;
    switch (tok.typ) {
    case plussym:
	eat(plussym);
	AST *exp = parseTerm();
	return ast_op_expr(opt, plusop, exp);
	break;
    case minussym:
	eat(minussym);
	AST *e = parseTerm();
	return ast_op_expr(opt, minusop, e);
	break;
    default:
	token_type expected[2] = {plussym, minussym};
	parse_error_unexpected(expected, 2, tok);
	break;
    }
    // The following should never execute
    return (AST *) NULL;
}

// <term> ::= <factor> { <mult-div-factor> }
// Return an AST with the operators (if any) associating to the left
static AST *parseTerm()
{
    token fst = tok;
    AST *fac = parseFactor();
    AST *exp = fac;
    while (tok.typ == multsym || tok.typ == divsym) {
	AST *rght = parseMultDivFactor();
	exp = ast_bin_expr(fst, exp, rght->data.op_expr.op,
			   rght->data.op_expr.exp);
    }
    return exp;
}

// <mult-div-factor> ::= <mult-div> <factor>
static AST *parseMultDivFactor()
{
    token opt = tok;
    switch (tok.typ) {
    case multsym:
	eat(multsym);
	AST *exp = parseFactor();
	return ast_op_expr(opt, multop, exp);
	break;
    case divsym:
	eat(divsym);
	AST *e = parseFactor();
	return ast_op_expr(opt, divop, e);
	break;
    default:
	token_type expected[2] = {multsym, divsym};
	parse_error_unexpected(expected, 2, tok);
	break;
    }
    // The following should never execute
    return (AST *) NULL;
}

// <factor> ::= <ident> | <paren-expr> | <signed-number>
static AST *parseFactor()
{
    switch (tok.typ) {
    case identsym:
	return parseIdentExpr();
	break;
    case lparensym:
	return parseParenExpr();
	break;
    case numbersym:
	return parseNumberExpr();
	break;
    default:
	token_type expected[3] =
	    {identsym, lparensym, numbersym};
	parse_error_unexpected(expected, 3, tok);
	break;
    }
    // The following should never execute
    return (AST *)NULL;
}

// Return an AST for an identifier use
static AST *parseIdentExpr()
{
    token idt = tok;
    eat(identsym);
    return ast_ident(idt, idt.text);
}

// Return an AST for a numeric literal
static AST *parseNumberExpr()
{
    token numt = tok;
    eat(numbersym);
    return ast_number(numt, word_as_float(numt.value));
}

// <paren-expr> ::= ( <expr> )
static AST *parseParenExpr()
{
    token lpt = tok;
    eat(lparensym);
    AST *ret = parseExpr();
    eat(rparensym);
    ret->file_loc = token2file_loc(lpt);
    return ret;
}
