/* $Id: unparser.c,v 1.9 2023/03/29 20:46:56 leavens Exp $ */
#include <stdio.h>
#include "ast.h"
#include "utilities.h"
#include "unparser.h"

// Amount of spaces to indent per nesting level
#define SPACES_PER_LEVEL 2

// Print SPACES_PER_LEVEL * level spaces to out
static void indent(FILE *out, int level)
{
    int num = SPACES_PER_LEVEL * level;
    for (int i = 0; i < num; i++) {
	fprintf(out, " ");
    }
}

// Unparse the given program AST with output going to the file out
void unparseProgram(FILE *out, AST *ast)
{
    AST_list vds = ast->data.program.vds;
    AST *stmt = ast->data.program.stmt;
    unparseVarDecls(out, vds, 0);
    unparseStmt(out, stmt, 0);
}

// Unparse the AST_list of var-decls to out
// with the given nesting level
// (note that if ast_list_is_empty(vds), then nothing is printed)
void unparseVarDecls(FILE *out, AST_list vds, int level)
{
    while (!ast_list_is_empty(vds)) {
	unparseVarDecl(out, ast_list_first(vds), level);
	vds = ast_list_rest(vds);
    }
}

// Unparse a single var-decl given by the AST vd to out,
// indented for the given nesting level
void unparseVarDecl(FILE *out, AST *vd, int level)
{
    indent(out, level);
    fprintf(out, "%s ", vt2str(vd->data.var_decl.vt));
    fprintf(out, "%s;\n", vd->data.var_decl.name);
}

// Unparse the statement given by the AST stmt to out,
// indented for the given level.
void unparseStmt(FILE *out, AST *stmt, int indentLevel)
{
    switch (stmt->type_tag) {
    case assign_ast:
	unparseAssignStmt(out, stmt, indentLevel);
	break;
    case begin_ast:
	unparseBeginStmt(out, stmt, indentLevel);
	break;
    case if_ast:
	unparseIfStmt(out, stmt, indentLevel);
	break;
    case read_ast:
	unparseReadStmt(out, stmt, indentLevel);
	break;
    case write_ast:
	unparseWriteStmt(out, stmt, indentLevel);
	break;
    default:
	bail_with_error("Call to unparseStmt with an AST that is not a statement!");
	break;
    }
}

// print a semicolon followed by a \n on the file out
static void semiAndNewline(FILE *out)
{
    fprintf(out, ";\n");
}

// Unparse the assignment statment given by stmt to out
// with indentation level given by level,
void unparseAssignStmt(FILE *out, AST *stmt, int level)
{
    indent(out, level);
    fprintf(out, "%s = ", stmt->data.assign_stmt.ident->data.ident.name);
    unparseExpr(out, stmt->data.assign_stmt.exp);
    semiAndNewline(out);
}

// Unparse the sequential statment given by stmt to out
// with indentation level given by level (indenting the body one more level)
void unparseBeginStmt(FILE *out, AST *stmt, int level)
{
    indent(out, level);
    fprintf(out, "{\n");
    unparseVarDecls(out, stmt->data.begin_stmt.vds, level+1);
    AST *stmt1 = ast_list_first(stmt->data.begin_stmt.stmts);
    AST_list rest = ast_list_rest(stmt->data.begin_stmt.stmts);
    unparseStmtList(out, stmt1, rest, level+1);
    indent(out, level);
    fprintf(out, "}\n");
}

// Unparse the list of statments given by stmt to out
// with indentation level given by level.
void unparseStmtList(FILE *out, AST *stmt1, AST_list rest,
			    int level)
{
    unparseStmt(out, stmt1, level);
    while (!ast_list_is_empty(rest)) {
	unparseStmt(out, ast_list_first(rest), level);
	rest = ast_list_rest(rest);
    }
}

// Unparse the if-statment given by stmt to out
// with indentation level given by level (and each body indented one more)
void unparseIfStmt(FILE *out, AST *stmt, int level)
{
    indent(out, level);
    fprintf(out, "if ( ");
    unparseExpr(out, stmt->data.if_stmt.exp);
    fprintf(out, " )\n");
    unparseStmt(out, stmt->data.if_stmt.stmt, level+1);
}

// Unparse the read statment given by stmt to out
void unparseReadStmt(FILE *out, AST *stmt, int level)
{
    indent(out, level);
    fprintf(out, "read %s", stmt->data.read_stmt.ident->data.ident.name);
    semiAndNewline(out);
}

// Unparse the write statment given by stmt to out.
void unparseWriteStmt(FILE *out, AST *stmt, int level)
{
    indent(out, level);
    fprintf(out, "write ");
    unparseExpr(out, stmt->data.write_stmt.exp);
    semiAndNewline(out);
}

// Unparse the expression given by the AST exp to out
// adding parentheses to indicate the nesting relationships
void unparseExpr(FILE *out, AST *exp)
{
    switch (exp->type_tag) {
    case bin_expr_ast:
	unparseBinExpr(out, exp);
	break;
    case ident_ast:
	unparseIdent(out, exp);
	break;
    case number_ast:
	unparseNumber(out, exp);
	break;
    case not_expr_ast:
	unparseNotExpr(out, exp);
	break;
    default:
	bail_with_error("Unexpected type_tag %d in unparseExpr", exp->type_tag);
	break;
    }
}

// Unparse the expression given by the AST exp to out
// adding parentheses (whether needed or not)
void unparseBinExpr(FILE *out, AST *exp)
{
    fprintf(out, "(");
    unparseExpr(out, exp->data.bin_expr.leftexp);
    fprintf(out, " ");
    fprintf(out, "%s", oper2str(exp->data.bin_expr.op));
    fprintf(out, " ");
    unparseExpr(out, exp->data.bin_expr.rightexp);
    fprintf(out, ")");
}

// Unparse the given identifer reference (use) to out
void unparseIdent(FILE *out, AST *id)
{
    fprintf(out, "%s", id->data.ident.name);
}

// Unparse the given number to out in decimal format
void unparseNumber(FILE *out, AST *num)
{
    word_print(out, num->data.number.value);
}

// Unparse the given logical negation expression to out
void unparseNotExpr(FILE *out, AST *exp)
{
    fprintf(out, "(!");
    unparseExpr(out, exp->data.not_expr.exp);
    fprintf(out, ")");
}
